"""Tests for the results module."""
from alglab.results import Results
import alglab.algorithm
import alglab.experiment
from sklearn.cluster import KMeans, SpectralClustering
import numpy as np


# We will use this KMeans implementation throughout the tests.
def kmeans(data: alglab.dataset.PointCloudDataset, k=10):
    sklearn_km = KMeans(n_clusters=k)
    sklearn_km.fit(data.data)
    return sklearn_km.labels_


def sc(data: alglab.dataset.PointCloudDataset, k=10):
    sklearn_sc = SpectralClustering(n_clusters=k)
    sklearn_sc.fit(data.data)
    return sklearn_sc.labels_


def test_plots():
    # Run a simple experiment
    results = Results("results/results.csv")
    assert results.num_runs == 2

    results.line_plot('noise', 'running_time_s')


def test_plots_multiple_parameters():
    alg1 = alglab.algorithm.Algorithm(kmeans)
    alg2 = alglab.algorithm.Algorithm(sc)

    noise_parameters = np.linspace(0, 1, 5)
    experiments = alglab.experiment.ExperimentalSuite(
        [alg1, alg2],
        alglab.dataset.TwoMoonsDataset,
        "results/twomoonsresults.csv",
        parameters={'k': 2,
                    'dataset.noise': noise_parameters,
                    'dataset.n': np.linspace(100, 1000, 3).astype(int)},
        evaluators=[alglab.evaluation.adjusted_rand_index]
    )
    results = experiments.run_all()
    results.line_plot('n', 'running_time_s', fixed_parameters={'noise': noise_parameters[0]})





"""
Create a generic class representing an algorithm which can be applied to a dataset.
"""
from typing import List, Type, Tuple, Callable, Dict, Union, get_type_hints
import inspect
import alglab.dataset
import time


class AlgorithmStep(object):
    def __init__(self,
                 name: str,
                 implementation: Callable,
                 first_step: bool):
        self.implementation = implementation
        self.name = name
        self.first_step = first_step

        # Automatically infer the return type
        self.return_type = self.__get_return_type()

        # Automatically infer the dataset class
        self.dataset_class = self.__get_dataset_class()

        # Automatically infer the parameter names for this algorithm step
        self.parameter_names = self.__get_parameter_names()

    def __get_return_type(self):
        sig = inspect.signature(self.implementation)
        if sig.return_annotation is not inspect._empty:
            return sig.return_annotation
        else:
            return object

    def __get_dataset_class(self):
        # Automatically infer the dataset class from type annotations
        sig = inspect.signature(self.implementation)

        non_default_parameters = [name for name, param in sig.parameters.items()
                                  if param.default == inspect.Parameter.empty]

        # Get the name of the dataset parameter
        if len(non_default_parameters) == 0:
            return alglab.dataset.NoDataset

        dataset_parameter = non_default_parameters[0]
        if not self.first_step:
            if len(non_default_parameters) == 0:
                raise ValueError("Second and later algorithm steps must take the output of the previous step as"
                                 " their first argument.")
            if len(non_default_parameters) == 1:
                return alglab.dataset.NoDataset
            if len(non_default_parameters) > 2:
                raise ValueError("All algorithm parameters must have default values.")
            dataset_parameter = non_default_parameters[1]
        else:
            if len(non_default_parameters) > 1:
                raise ValueError("All algorithm parameters must have default values.")

        # Extract the type hint for the dataset parameter, if one exists
        type_hints = get_type_hints(self.implementation)
        if dataset_parameter in type_hints:
            return type_hints[dataset_parameter]
        else:
            return alglab.dataset.Dataset

    def __get_parameter_names(self):
        inferred_parameter_names = []
        sig = inspect.signature(self.implementation)
        for name, param in sig.parameters.items():
            if param.default != inspect.Parameter.empty:
                if name not in inferred_parameter_names:
                    inferred_parameter_names.append(name)
        return inferred_parameter_names

    def run(self, dataset: alglab.dataset.Dataset, params: Dict, previous_step_output=None):
        if not isinstance(dataset, self.dataset_class):
            raise TypeError("Provided dataset type must match dataset_class expected by the implementation.")

        # Filter the parameters by the ones accepted by this step
        this_step_parameters = {k: v for k, v in params.items() if k in self.parameter_names}

        non_kw_args = []
        if not self.first_step:
            non_kw_args.append(previous_step_output)
        if self.dataset_class is not alglab.dataset.NoDataset:
            non_kw_args.append(dataset)

        start_time = time.time()
        result = self.implementation(*non_kw_args, **this_step_parameters)
        end_time = time.time()
        running_time = end_time - start_time

        return result, running_time


class Algorithm(object):

    def __init__(self,
                 implementation: Union[Callable, List[Callable], List[Tuple[str, Callable]]],
                 name: str = None):
        """Create an algorithm definition. The implementation should be a python method which takes
        a dataset as a positional argument (if dataset_class is not NoDataset) and
        the parameters as keyword arguments. The implementation should return an object of type
        return_type.
        """
        self.implementation = []
        if isinstance(implementation, Callable):
            self.implementation = [AlgorithmStep(implementation.__name__, implementation, True)]
        elif isinstance(implementation, List):
            if len(implementation) == 0:
                raise ValueError("Algorithm must have at least one step.")
            if isinstance(implementation[0], Callable):
                first_step = True
                for imp in implementation:
                    self.implementation.append(AlgorithmStep(imp.__name__, imp, first_step))
                    first_step = False
            elif isinstance(implementation[0], Tuple):
                first_step = True
                for imp in implementation:
                    assert isinstance(imp, Tuple)
                    if len(imp) != 2:
                        raise ValueError("Algorithm steps should be tuples with the step name and implementation.")
                    self.implementation.append(AlgorithmStep(imp[0], imp[1], first_step))
                    first_step = False
        self.name = name if name is not None else self.implementation[0].name

        if self.name is "dataset":
            raise ValueError("It is not permitted to call an algorithm 'dataset'.")

        # Check for a return type hint
        self.return_type = self.implementation[-1].return_type

        # Automatically infer the dataset class if it is not provided
        self.dataset_class = self.implementation[0].dataset_class

        # Automatically infer the parameter names if they are not provided
        self.all_parameter_names = self.__get_parameters()

        # Check that the number of non-defaulted parameters in each step of the implementation is correct.
        self.__check_number_of_parameters()

        self.time_headings = [f'{step.name}_running_time_s' for step in self.implementation]
        self.time_headings.append('running_time_s')

    def __get_parameters(self):
        inferred_parameter_names = []
        for step in self.implementation:
            for par_name in step.parameter_names:
                if par_name not in inferred_parameter_names:
                    inferred_parameter_names.append(par_name)
        return inferred_parameter_names

    def __check_number_of_parameters(self):
        for step in self.implementation:
            if ((step.dataset_class is alglab.dataset.NoDataset and self.dataset_class is not alglab.dataset.NoDataset) or
                    (step.dataset_class is not alglab.dataset.NoDataset and self.dataset_class is alglab.dataset.NoDataset)):
                raise ValueError("All algorithm steps must take the dataset as an argument.")


    def run(self, dataset: alglab.dataset.Dataset, params: Dict):
        if not isinstance(dataset, self.dataset_class):
            raise TypeError("Provided dataset type must match dataset_class expected by the implementation.")

        for param in params.keys():
            if param not in self.all_parameter_names:
                raise ValueError("Unexpected parameter name.")

        result = None
        running_times = {}
        start_time = time.time()
        for step in self.implementation:
            result, running_time = step.run(dataset, params, previous_step_output=result)
            running_times[f'{step.name}_running_time_s'] = running_time
        end_time = time.time()
        running_times['running_time_s'] = end_time - start_time

        if not isinstance(result, self.return_type):
            raise TypeError("Provided result type must match promised return_type.")

        return result, running_times

    def __repr__(self):
        return self.name



"""Tests for the experiment module."""
from sklearn.cluster import KMeans, SpectralClustering
import numpy as np
import pytest
import alglab


# We will use this KMeans implementation throughout the tests.
def kmeans(data: alglab.dataset.PointCloudDataset, k=10):
    sklearn_km = KMeans(n_clusters=k)
    sklearn_km.fit(data.data)
    return sklearn_km.labels_


def sc(data: alglab.dataset.PointCloudDataset, k=10):
    sklearn_sc = SpectralClustering(n_clusters=k)
    sklearn_sc.fit(data.data)
    return sklearn_sc.labels_


def test_experimental_suite():
    # Test the experimental suite class as it's intended to be used.
    alg1 = alglab.algorithm.Algorithm(kmeans)
    alg2 = alglab.algorithm.Algorithm(sc)

    experiments = alglab.experiment.ExperimentalSuite(
        [alg1, alg2],
        alglab.dataset.TwoMoonsDataset,
        "results/twomoonsresults.csv",
        parameters={
            "kmeans.k": 2,
            "sc.k": 2,
            "dataset.n": 1000,
            "dataset.noise": np.linspace(0, 1, 5),
        },
        evaluators=[alglab.evaluation.adjusted_rand_index]
        )
    experiments.run_all()


def test_multiple_runs():
    # Test the experimental suite class as it's intended to be used.
    alg1 = alglab.algorithm.Algorithm(kmeans)
    alg2 = alglab.algorithm.Algorithm(sc)

    experiments = alglab.experiment.ExperimentalSuite(
        [alg1, alg2],
        alglab.dataset.TwoMoonsDataset,
        "results/twomoonsresults.csv",
        parameters={
            "kmeans.k": 2,
            "sc.k": 2,
            "dataset.n": 1000,
            "dataset.noise": np.linspace(0, 1, 5),
        },
        evaluators=[alglab.evaluation.adjusted_rand_index],
        num_runs=2
    )
    assert experiments.num_trials == 20
    experiments.run_all()


def test_dynamic_params():
    alg1 = alglab.algorithm.Algorithm(kmeans)
    alg2 = alglab.algorithm.Algorithm(sc)

    experiments = alglab.experiment.ExperimentalSuite(
        [alg1, alg2],
        alglab.dataset.TwoMoonsDataset,
        "results/twomoonsresults.csv",
        parameters={
            "kmeans.k": 2,
            "sc.k": [(lambda p: int(p['n'] / 100)), 2],
            "dataset.noise": 0.1,
            "dataset.n": np.linspace(100, 1000, 5).astype(int),
        },
        evaluators=[alglab.evaluation.adjusted_rand_index]
    )
    experiments.run_all()


def test_simple_configuration():
    experiments = alglab.experiment.ExperimentalSuite([kmeans, sc],
                                                      alglab.dataset.TwoMoonsDataset,
                                                      "results/twomoonsresults.csv")
    experiments.run_all()


def test_simple_with_custom_evaluator():
    def const_evaluator(data, alg_output):
        return 5

    experiments = alglab.experiment.ExperimentalSuite([kmeans, sc],
                                                      alglab.dataset.TwoMoonsDataset,
                                                      "results/twomoonsresults.csv",
                                                      evaluators=[const_evaluator])
    experiments.run_all()


def test_wrong_alg_name():
    algs = [alglab.algorithm.Algorithm(kmeans),
            alglab.algorithm.Algorithm(sc)]

    with pytest.raises(ValueError, match='algorithm'):
        experiments = alglab.experiment.ExperimentalSuite(
            algs,
            alglab.dataset.TwoMoonsDataset,
            "results/twomoonsresults.csv",
            parameters={
                "spectral_clustering.k": 2,
                "dataset.n": 1000,
                "dataset.noise": np.linspace(0, 1, 5),
            },
            evaluators=[alglab.evaluation.adjusted_rand_index]
        )


def test_multi_step_algs():
    def kmeans_init(data: alglab.dataset.PointCloudDataset, k=10):
        return KMeans(n_clusters=k)

    def kmeans_predict(kmeans_obj, data: alglab.dataset.PointCloudDataset) -> np.ndarray:
        kmeans_obj.fit(data.data)
        return kmeans_obj.labels_

    def kmeans_slow_predict(kmeans_obj, data: alglab.dataset.PointCloudDataset) -> np.ndarray:
        kmeans_obj.fit(data.data)
        kmeans_obj.fit(data.data)
        return kmeans_obj.labels_

    alg1 = alglab.algorithm.Algorithm([('fit', kmeans_init), ('predict', kmeans_predict)], name='kmeans')
    alg2 = alglab.algorithm.Algorithm([('fit', kmeans_init), ('predict', kmeans_slow_predict)], name='slow_kmeans')

    experiments = alglab.experiment.ExperimentalSuite(
        [alg1, alg2],
        alglab.dataset.TwoMoonsDataset,
        "results/twomoonsresults.csv",
        parameters={
            "k": 2,
            "dataset.n": np.linspace(100, 1000, 6),
        },
    )

    results = experiments.run_all()
    results.line_plot("n", "fit_running_time_s")
    results.line_plot("n", "predict_running_time_s")
    results.line_plot("n", "running_time_s")



"""Tests for the algorithm class."""
import pytest

import alglab.dataset
import alglab.algorithm
import alglab.evaluation
from sklearn.cluster import KMeans
import numpy as np
from typing import List


def kmeans(data: alglab.dataset.PointCloudDataset, k=10) -> np.ndarray:
    sklearn_km = KMeans(n_clusters=k)
    sklearn_km.fit(data.data)
    return sklearn_km.labels_


def kmeans_no_return(data: alglab.dataset.PointCloudDataset, k=10):
    sklearn_km = KMeans(n_clusters=k)
    sklearn_km.fit(data.data)
    return sklearn_km.labels_


def no_dataset_impl(n=10):
    return n ** 3


def test_algorithm():
    # Test the algorithm class as it is intended to be used.
    kmeans_algorithm = alglab.algorithm.Algorithm(kmeans,
                                                  name='kmeans')

    test_data = alglab.dataset.TwoMoonsDataset()
    _ = kmeans_algorithm.run(test_data, {'k': 8})


def test_fit_predict_algorithm():
    def kmeans_init(data: alglab.dataset.PointCloudDataset, k=10):
        return KMeans(n_clusters=k)

    def kmeans_predict(kmeans_obj, data: alglab.dataset.PointCloudDataset) -> np.ndarray:
        kmeans_obj.fit(data.data)
        return kmeans_obj.labels_

    kmeans_algorithm = alglab.algorithm.Algorithm([kmeans_init, kmeans_predict],
                                                  name='kmeans')

    test_data = alglab.dataset.TwoMoonsDataset()
    labels, times = kmeans_algorithm.run(test_data, {'k': 8})
    assert(len(times.keys()) == 3)


def test_no_return_type():
    kmeans_algorithm = alglab.algorithm.Algorithm(kmeans_no_return)
    test_data = alglab.dataset.TwoMoonsDataset()
    _ = kmeans_algorithm.run(test_data, {'k': 8})


def test_bad_specified_return_type():
    # Create an implementation with the wrong return type hint
    def bad_return(n=10) -> np.ndarray:
        return n ** 3

    alg = alglab.algorithm.Algorithm(bad_return)

    with pytest.raises(TypeError, match='return'):
        _ = alg.run(alglab.dataset.NoDataset(), {'n': 20})


def test_infer_parameters_with_dataset():
    kmeans_algorithm = alglab.algorithm.Algorithm(kmeans)

    test_data = alglab.dataset.TwoMoonsDataset()
    _ = kmeans_algorithm.run(test_data, {'k': 8})

    with pytest.raises(ValueError, match='parameter'):
        _ = kmeans_algorithm.run(test_data, {'data': test_data})


def test_all_params_default():
    def non_defaults(k, l, n=10) -> int:
        return k * l * n ** 3

    with pytest.raises(ValueError, match='default'):
        alg = alglab.algorithm.Algorithm(non_defaults)


def test_pass_wrong_dataset_type():
    kmeans_algorithm = alglab.algorithm.Algorithm(kmeans)

    test_data = alglab.dataset.GraphDataset()
    with pytest.raises(TypeError, match='dataset'):
        _ = kmeans_algorithm.run(test_data, {'k': 8})


def test_no_dataset_type_hint():
    def kmeans_nohint(data, k=10):
        sklearn_km = KMeans(n_clusters=k)
        sklearn_km.fit(data.data)
        return sklearn_km.labels_

    alg = alglab.algorithm.Algorithm(kmeans_nohint)
    test_data = alglab.dataset.TwoMoonsDataset()
    _ = alg.run(test_data, {'k': 8})


def test_all_type_hints():
    def kmeans_hints(data: alglab.dataset.PointCloudDataset, k: int = 10):
        sklearn_km = KMeans(n_clusters=k)
        sklearn_km.fit(data.data)
        return sklearn_km.labels_

    alg = alglab.algorithm.Algorithm(kmeans_hints)
    test_data = alglab.dataset.TwoMoonsDataset()
    _ = alg.run(test_data, {'k': 8})


def test_automatically_infer_parameters_bad():
    def good_return(n=10) -> int:
        return n ** 3

    alg = alglab.algorithm.Algorithm(good_return)

    with pytest.raises(ValueError, match='parameter'):
        _ = alg.run(alglab.dataset.NoDataset(), {'b': 20})


def test_unspecified_parameter():
    # Test the algorithm class as it is intended to be used.
    kmeans_algorithm = alglab.algorithm.Algorithm(kmeans)

    # Not fully specifying the parameter is permitted when running.
    test_data = alglab.dataset.TwoMoonsDataset()
    _ = kmeans_algorithm.run(test_data, {})


def test_no_dataset():
    cube_algorithm = alglab.algorithm.Algorithm(no_dataset_impl,
                                                name="power")

    _ = cube_algorithm.run(alglab.dataset.NoDataset(), {'n': 20})


def test_wrong_dataset():
    cube_algorithm = alglab.algorithm.Algorithm(no_dataset_impl,
                                                name="power")

    with pytest.raises(TypeError, match='dataset'):
        _ = cube_algorithm.run(alglab.dataset.TwoMoonsDataset(), {'n': 20})


def test_openml_dataset():
    dataset = alglab.dataset.OpenMLDataset(name="iris")
    kmeans_algorithm = alglab.algorithm.Algorithm(kmeans)

    labels, _ = kmeans_algorithm.run(dataset, {'k': 3})

    # Check that evaluation works
    _ = alglab.evaluation.adjusted_rand_index.apply(dataset, labels)



"""
Methods for evaluating the performance of an algorithm.
"""
from typing import Callable, Type, get_type_hints
import numpy as np
import alglab
import stag.cluster
import stag.graph
import scipy.sparse.linalg
import inspect


class Evaluator(object):

    def __init__(self,
                 implementation: Callable,
                 name: str = None,
                 alg_result_type: Type = None,
                 dataset_class: Type[alglab.dataset.Dataset] = None):
        """Define a method of evaluating an algorithm. Specify the evaluator implementation as
        well as the expected result type of the algorithm to be evaluated and the type of the dataset
        the algorithm should be applied to."""
        self.implementation = implementation
        sig = inspect.signature(self.implementation)

        self.name = name if name is not None else implementation.__name__

        # Automatically infer the dataset class from type annotations
        parameter_names = [
            name for name, param in sig.parameters.items()
        ]

        if len(parameter_names) < 1:
            raise ValueError("Evaluator implementation must take at least one parameter.")
        if len(parameter_names) > 2:
            raise ValueError("Evaluator implementation must take at most two parameters: (optionally) a dataset, "
                             "and the algorithm's output.")

        self.dataset_class = dataset_class
        if self.dataset_class is None:
            # If there is only one positional argument to the implementation, then there is no dataset.
            if len(parameter_names) == 1:
                self.dataset_class = alglab.dataset.NoDataset
            else:
                # Get the name of the dataset parameter
                dataset_parameter = parameter_names[0]

                # Extract the type hint for the dataset parameter, if one exists
                type_hints = get_type_hints(implementation)
                if dataset_parameter in type_hints:
                    self.dataset_class = type_hints[dataset_parameter]
                else:
                    self.dataset_class = alglab.dataset.Dataset

        self.alg_result_type = alg_result_type
        if self.alg_result_type is None:
            # Get the name of the result parameter
            result_parameter = parameter_names[-1]

            # Extract the type hint for the result parameter, if it exists
            type_hints = get_type_hints(implementation)
            if result_parameter in type_hints:
                self.alg_result_type = type_hints[result_parameter]
            else:
                self.alg_result_type = object

    def apply(self, dataset: alglab.dataset.Dataset, alg_result):
        if not isinstance(dataset, self.dataset_class):
            raise TypeError(f"Expected dataset type to be {self.dataset_class}, got {type(dataset)}.")

        if not isinstance(alg_result, self.alg_result_type):
            raise TypeError(f"Expected alg_result type to be {self.alg_result_type} but got {type(alg_result)}.")

        if self.dataset_class is not alglab.dataset.NoDataset:
            result = self.implementation(dataset, alg_result)
        else:
            result = self.implementation(alg_result)

        return result

    def __str__(self):
        return self.name

# -----------------------------------------------------------------------------
# Clustering Evaluation
# -----------------------------------------------------------------------------


def __ari_impl(data: alglab.dataset.ClusterableDataset, labels):
    if data.gt_labels is not None:
        return stag.cluster.adjusted_rand_index(data.gt_labels, labels)
    else:
        raise ValueError('No ground truth labels provided.')


adjusted_rand_index = Evaluator(__ari_impl,
                                name="adjusted_rand_index",
                                alg_result_type=np.ndarray,
                                dataset_class=alglab.dataset.ClusterableDataset)


# -----------------------------------------------------------------------------
# Graph Evaluation
# -----------------------------------------------------------------------------

def __num_vertices_impl(_: alglab.dataset.Dataset, graph: stag.graph.Graph):
    return graph.number_of_vertices()


num_vertices = Evaluator(__num_vertices_impl,
                         name="number_of_vertices",
                         alg_result_type=stag.graph.Graph,
                         dataset_class=alglab.dataset.Dataset)


def __avg_degree_impl(_: alglab.dataset.Dataset, graph: stag.graph.Graph):
    return graph.average_degree()


avg_degree = Evaluator(__avg_degree_impl,
                       name="average_degree",
                       alg_result_type=stag.graph.Graph,
                       dataset_class=alglab.dataset.Dataset)


def __normalised_laplacian_second_eigenvalue_impl(_: alglab.dataset.Dataset, graph: stag.graph.Graph):
    lap = graph.normalised_laplacian().to_scipy()
    eigs, _ = scipy.sparse.linalg.eigsh(lap, which='SM', k=2)
    return eigs[1]


normalised_laplacian_second_eigenvalue = Evaluator(__normalised_laplacian_second_eigenvalue_impl,
                                                   name="lap_second_eigenvalue",
                                                   alg_result_type=stag.graph.Graph,
                                                   dataset_class=alglab.dataset.Dataset)
